from copy import deepcopy
from operator import attrgetter
from random import Random
from typing import Callable

from sampo.generator import SimpleSynthetic
from sampo.generator.types import SyntheticGraphType
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.utilities.collections import build_index


class BlockNode:
    """
    `BlockNode` represents the node of `BlockGraph` and contains corresponding
    `WorkGraph` and related dependencies between blocks in `BlockGraph`.
    """
    def __init__(self, wg: WorkGraph):
        self.wg = wg
        self.blocks_from: list[BlockNode] = []
        self.blocks_to: list[BlockNode] = []

    @property
    def id(self):
        return self.wg.start.id


class BlockGraph:
    """
    Represents the block graph, where blocks are instances of `WorkGraph` and
    edges are simple *FS* dependencies.
    """
    def __init__(self, nodes: list[WorkGraph]):
        self.nodes = [BlockNode(node) for node in nodes]
        self.node_dict = build_index(self.nodes, attrgetter('id'))

    def __getitem__(self, item) -> BlockNode:
        return self.node_dict[item]

    def to_work_graph(self) -> WorkGraph:
        """
        Creates `WorkGraph` that are equal to this `BlockGraph`.
        """
        copied_nodes = deepcopy(self.nodes)
        global_start: GraphNode = [node for node in copied_nodes if len(node.blocks_from) == 0][0].wg.start
        global_end:   GraphNode = [node for node in copied_nodes if len(node.blocks_to) == 0][0].wg.finish

        for end in copied_nodes:
            end.wg.start.add_parents([start.wg.finish for start in end.blocks_from])

        return WorkGraph(global_start, global_end)

    @staticmethod
    def add_edge(start: BlockNode, end: BlockNode):
        start.blocks_to.append(end)
        end.blocks_from.append(start)


def generate_blocks(n_blocks: int, type_prop: list[int],
                    count_supplier: Callable[[int], tuple[int, int]],
                    edge_prob: float, rand: Random | None = Random()) -> BlockGraph:
    """
    Generate block graph according to given parameters.

    :param n_blocks: the count of blocks
    :param type_prop: proportions of the `WorkGraph` types: General, Parallel, Sequential
    :param count_supplier: function that computes the borders of block size from it's index
    :param edge_prob: edge existence probability
    :param rand: a random reference
    :return: generated block graph
    """
    ss = SimpleSynthetic(rand)

    modes = rand.sample(list(SyntheticGraphType), counts=[p * n_blocks for p in type_prop], k=n_blocks)
    nodes = [ss.work_graph(mode, *count_supplier(i)) for i, mode in enumerate(modes)]
    bg = BlockGraph(nodes)

    rev_edge_prob = int(1 / edge_prob)
    for idx, start in enumerate(bg.nodes):
        for end in bg.nodes[idx:]:
            if start == end or rand.randint(0, rev_edge_prob) != 0:
                continue
            bg.add_edge(start, end)

    return bg



from copy import deepcopy
from typing import Iterable

from sampo.scheduler.utils.block_graph import BlockGraph
from sampo.scheduler.utils.multi_agency import ScheduledBlock
from sampo.schemas.contractor import Contractor
from sampo.utilities.validation import validate_schedule, \
    check_all_allocated_workers_do_not_exceed_capacity_of_contractors


def validate_block_schedule(bg: BlockGraph, schedule: dict[str, ScheduledBlock]):
    _check_block_dependencies(bg, schedule)
    _check_blocks_separately(schedule.values())

    contractors = set([contractor for sblock in schedule.values() for contractor in sblock.agent.contractors])
    _check_blocks_with_global_timelines(schedule.values(), contractors)


def _check_block_dependencies(bg: BlockGraph, schedule: dict[str, ScheduledBlock]):
    for sblock in schedule.values():
        for p in bg[sblock.id].blocks_from:
            parent_sblock = schedule[p.id]
            assert parent_sblock.start_time <= parent_sblock.end_time <= sblock.start_time <= sblock.end_time


def _check_blocks_separately(sblocks: Iterable[ScheduledBlock]):
    for sblock in sblocks:
        validate_schedule(sblock.schedule, sblock.wg, sblock.agent.contractors)


def _check_blocks_with_global_timelines(sblocks: Iterable[ScheduledBlock], contractors: Iterable[Contractor]):
    """
    Checks that no agent's contractor uses more resources that can supply.

    Note that this should fail if there is a shared contractor between agents, but this
    term is, of course, unsupported.

    :param sblocks: scheduled blocks of works
    :param contractors: global scope of contractors(collected from all agents used to construct sblocks)
    """
    initial_worker_pool: dict[str, dict[str, int]] = {}
    for contractor in contractors:
        initial_worker_pool[contractor.id] = {}
        for w in contractor.workers.values():
            if w.name in initial_worker_pool[contractor.id].keys():
                initial_worker_pool[contractor.id][w.name] += w.count
            else:
                initial_worker_pool[contractor.id][w.name] = w.count
    cur_worker_pool = deepcopy(initial_worker_pool)

    for sblock in sblocks:
        check_all_allocated_workers_do_not_exceed_capacity_of_contractors(sblock.schedule,
                                                                          initial_worker_pool,
                                                                          cur_worker_pool)



import random
from typing import Optional, List

from sampo.schemas.requirements import WorkerReq
from sampo.utilities.sampler.types import MinMax
from sampo.utilities.sampler.resources import WORKER_TYPES, WorkerSpecialization


def get_worker_req(rand: random.Random,
                   name: str,
                   volume: Optional[MinMax[int]] = MinMax[int](1, 50),
                   worker_count: Optional[MinMax[int]] = MinMax[int](1, 100)
                   ) -> WorkerReq:
    assert type(rand) == random.Random
    assert type(name) == str
    count = rand.randint(volume.min, volume.max)
    return WorkerReq(name, count, worker_count.min, worker_count.max)


def get_worker_reqs_list(rand: random.Random,
                         volume: Optional[MinMax[int]] = MinMax[int](1, 50),
                         worker_count: Optional[MinMax[int]] = MinMax[int](1, 100)
                         ) -> List[WorkerReq]:
    assert type(rand) == random.Random
    names: List[WorkerSpecialization] = list(WORKER_TYPES)
    rand.shuffle(names)
    req_count = rand.randint(1, len(names))
    names = names[:req_count]
    return get_worker_specific_reqs_list(rand, names, volume, worker_count)


def get_worker_specific_reqs_list(rand: random.Random,
                                  worker_names: List[WorkerSpecialization],
                                  volume: Optional[MinMax[int]] = MinMax[int](1, 50),
                                  worker_count: Optional[MinMax[int]] = MinMax[int](1, 100)
                                  ) -> List[WorkerReq]:
    assert type(rand) == random.Random
    reqs = [get_worker_req(rand, name, volume, worker_count) for name in worker_names]
    return reqs
